MODULE FIQMachine4;

IMPORT HelyOS, HKernel, HLog, FPE, SYSTEM;

CONST
    (* SYSTEM DEPENDENT DEEINITIONS *)
    (*OSDELAY =  18432*4;  20 ms *)
    OSDELAY =  3686;  (* 1ms *)
    (* TIMESTEP = 1000*OSDELAY DIV 3686400; *)
    TIMESTEP = 1;
    (* Register for Timer Interrupt *)
    OSCR = 90000010H;
    OSMR0 = 90000000H;
    OSSR = 90000014H;
    OSIER = 9000001CH;
    
    (* Register for Interrupt handling  *)
    ICIP = 90050000H;   (* IRQ pending register *)
    ICFP = 90050010H;   (* FIQ pending register *)
    ICMR = 90050004H;   (* int mask register *)
    ICLR = 90050008H;   (* route the int to IRQ or FIQ *)
    OSTimerBit = 4000000H; NOSTImerBit = 0FBFFFFFFH; (* 26 *)
    Uart3Bit = 20000H; NUart3Bit = 0FFFDFFFFH;  (* 17 *)
    Uart2Bit = 10000H; NUart2Bit = 0FFFEFFFFH; (* 16 *)
    Uart1Bit = 8000H; NUart1Bit = 0FFFF7FFFH; (* 15 *)
    FPGABit = 400H; NFPGABit = 0FFFFFBFFH; (* 10 *)
    ExtUARTBit = 200H; NExtUARTBit = 0FFFFFDFFH;(* 9 *)
    
    (* LED *)
        GPSR = 90040008H;
        GPCR = 9004000CH;
        LED0* = 00100000H;
        LED1* = 00200000H;
        LED2* = 00400000H;
        LED3* = 00800000H;(*  GPIO 23 *)
        
        LEDCONST = 200000;

(* Ranges *)

  MAXDRV*  = 10; (* reduced for memory usage *)
  MAXCND*  = 10; (* reduced for memory usage *)
  MAXTSK*  = 100;

  MAXTSKI* = 100;
  MAXTRGB* = 20;
  MAXTHIS* = 20;
  
  MAXPRG*  = 1000;
  MAXTRACE* = 500;

(* Task States *)

  COMPLETE* = 0;
  RELEASED* = 1;
  DISPATCHED* = 2;

(* Opcodes *)

  NOP* = 0;
  CALL* = 1;
  SCHEDULE* = 2;
  FUTURE* = 3;
  DISPATCH* = 4;
  IDLE* = 5;
  FORK* = 6;
  RET* = 7;
  JUMP* = 8;
  IFT* = 9;

TYPE
   
(* Drivers *)
   Driver* = PROCEDURE();
   Drivers* = ARRAY MAXDRV OF Driver;

(* Conditions *)
   Condition* = PROCEDURE(): BOOLEAN;
   Conditions* = ARRAY MAXCND OF Condition;
   
(* Tasks *)
   Task* = PROCEDURE();
   Tasks = ARRAY MAXTSK OF Task;

(* Trace : we assume a periodical system !!!*)
    Trace* = RECORD from*, to*, time*: INTEGER; c*: BOOLEAN END;
    Traces* = ARRAY MAXTRACE OF Trace;
     
(* Task Context *)
   
   TaskContext = RECORD
      optimization: INTEGER;
      sp, fpfiq, spfiq: INTEGER;
      spsr, spsrfiq: INTEGER; (* saved processor status in task mode *)
      stackfiq: ARRAY 1024 OF INTEGER; (* the fiq stack *)
      stack: ARRAY 1024 OF INTEGER; (* the task stack *)
   END;

(* Tasks Instance *)
   TaskInstance = RECORD
      taskcode : Task; (* task code *)
      state: INTEGER; (* COMPLETE/RELEASED/DISPATCHED *)
      deadline: INTEGER; (* absolute deadline *)
      context: TaskContext; (* task context MUST be in the last position of the RECORD !!!!!!*)
   END;
   TaskSet = ARRAY MAXTSKI OF TaskInstance;

(* Trigger Binding *)
   TriggerBinding = RECORD
      g: INTEGER; (* absolute state *)
      a: INTEGER; (* interrupt code address *)
   END;
   TriggerQueue = ARRAY MAXTRGB OF TriggerBinding;

(* Thread Set*)
   ThreadInstance = RECORD
      taskID: INTEGER; (* TaskID *)
      at: INTEGER; (* completion code address *)
      g: INTEGER; (* absolute state *)
      a: INTEGER; (* continuation code address *)
      s: INTEGER; (* reference state *)
   END;
   ThreadSet = ARRAY MAXTHIS OF ThreadInstance;

(* Instruction *)
   Instruction = RECORD
      opcode*: INTEGER;
      arg1*, arg2*, arg3*: INTEGER;
   END;
   
(* Program *)
   Program* = ARRAY MAXPRG OF Instruction;

VAR
   drivers: Drivers;
   NDRV: INTEGER; 

   conditions: Conditions;
   NCND: INTEGER;
   
   tasks*: Tasks;
   NTSK*: INTEGER;

   program*: Program;
   NINS*: INTEGER;
   
   traces*: Traces;
    traceidx: INTEGER;
   TIMEOVERFLOW: INTEGER;

(*********************************************************************)
(************************* SYSTEM MACHINE ****************************)
(*********************************************************************)

   logicaltime: INTEGER;
   NRT : INTEGER;

   taskset: TaskSet;
   NTKIS: INTEGER;

   triggerqueue: TriggerQueue;
   NTRGB: INTEGER;

   threadset: ThreadSet;
   NTHIS: INTEGER;

   yieldtotasks: BOOLEAN;

   pc: INTEGER;
   referencestate: INTEGER;
   opcode: INTEGER;
   branch: BOOLEAN;

   drivercode: Driver;
   conditioncode: Condition;

   runningtask: INTEGER;
   
   index: INTEGER;
   hit: INTEGER;
   
   localproc: Task;
   main: TaskContext; (* we discard the stack !!! *)

(* implementation dependent stuff *)
    sp, spfiq, fpfiq, spsr, spsrfiq: INTEGER; 
    icfp : INTEGER;
    taskID, at, deadline, g, a, s: INTEGER;
    
    (* profiler *)
    utilization : INTEGER;
    peak, delay, min: INTEGER;
    mean: INTEGER;
    delta: INTEGER;
    
    scode, ecode, edf: INTEGER;
    delta1, delta2: INTEGER;
    delaye, delays, delayedf: INTEGER;
    
    preemption : INTEGER;
    preempted: INTEGER;
    
    (* Errors *)
    DEADLINEOVR, THREADSETOVR, TRIGGEROVR, UNDEFINSTR: BOOLEAN;
    (* test *)
    LEDON: INTEGER;

PROCEDURE ResetErrors;
BEGIN
    DEADLINEOVR := FALSE;
    THREADSETOVR := FALSE;
    TRIGGEROVR := FALSE;
    UNDEFINSTR := FALSE;
END ResetErrors;

PROCEDURE ShowErrors();
BEGIN
    IF DEADLINEOVR THEN HLog.Str("Deadline overrun"); HLog.Ln; END;
    IF THREADSETOVR THEN HLog.Str("Thread set overflow"); HLog.Ln; END;
    IF TRIGGEROVR THEN HLog.Str("Trigger queue overflow"); HLog.Ln; END;
    IF UNDEFINSTR THEN HLog.Str("Undefined instruction"); HLog.Ln; END;
END ShowErrors;

PROCEDURE EnableInterrupt();
BEGIN
   SYSTEM.LDPSR(21H, HKernel.SVCMode); (* Enable all interrupts *)
END EnableInterrupt;

PROCEDURE DisableInterrupt();
BEGIN
   SYSTEM.LDPSR(21H, HKernel.SVCMode + HKernel.FIQDisabled + HKernel.IRQDisabled);  (* Disable interrupts *)
END DisableInterrupt;

PROCEDURE ReRouteOSTimer;
    VAR i:INTEGER;
BEGIN
    SYSTEM.LDPSR(21H, HKernel.SVCMode + HKernel.FIQDisabled + HKernel.IRQDisabled);  (* Disable interrupts *)
    (* re-route Timer to FIQ *)
    SYSTEM.PUT(ICLR, OSTimerBit); (* Route all int to IRQ, but Match0 on FIQ (bit 26) *)
    SYSTEM.GET(OSCR, i); SYSTEM.PUT(OSMR0, i + OSDELAY);
    SYSTEM.PUT(OSSR, 1); (* Clear Match *)
    SYSTEM.PUT(OSIER,1); (* Enable Interrupt on Match0 *)
    SYSTEM.GET(ICMR, i); 
    i := SYSTEM.ORR(i, OSTimerBit); 
    SYSTEM.PUT(ICMR, i); (* Enable Interrupt on OSTimer *)
    SYSTEM.LDPSR(21H, HKernel.SVCMode); (* Enable all interrupts *)
END ReRouteOSTimer;

PROCEDURE InitTaskSet;
BEGIN
   NRT := 0;
   index := 0;
   
   WHILE (index < NTSK) DO
      taskset[index].taskcode := tasks[index];
      taskset[index].state := COMPLETE;
      taskset[index].deadline := -1;

      INC(index)
   END;
END InitTaskSet;

(*
PROCEDURE CreateTaskInstance(taskID, deadline, s: INTEGER);
BEGIN
   IF taskset[taskID].state = COMPLETE THEN
      taskset[taskID].state := RELEASED;
      INC(NRT);
   ELSE DEADLINEOVR := TRUE;
      HLog.Str("Deadline overrun!"); HLog.Ln();
   END;

   taskset[taskID].deadline := deadline + s
END CreateTaskInstance;
*)

PROCEDURE (8) CreateTaskInstance2();
BEGIN
   IF taskset[taskID].state = COMPLETE THEN
      taskset[taskID].state := RELEASED;
      INC(NRT);
   ELSE DEADLINEOVR := TRUE;
   END;

   taskset[taskID].deadline := deadline + s
END CreateTaskInstance2;

PROCEDURE CreateTriggerBinding(g, a, s: INTEGER);
BEGIN
   IF NTRGB < MAXTRGB THEN
      triggerqueue[NTRGB].g := g + s;
      triggerqueue[NTRGB].a := a;

      INC(NTRGB)
   ELSE TRIGGEROVR := TRUE;
      HLog.Str("Trigger queue overflow!"); HLog.Ln();
   END;
END CreateTriggerBinding;

PROCEDURE (8) CreateTriggerBinding2();
BEGIN
   IF NTRGB < MAXTRGB THEN
      triggerqueue[NTRGB].g := g + s;
      triggerqueue[NTRGB].a := a;

      INC(NTRGB)
   ELSE TRIGGEROVR := TRUE;
   END;
END CreateTriggerBinding2;

(* PROCEDURE CreateThreadInstance(taskID, at, g, a, s: INTEGER);
BEGIN
   IF NTHIS < MAXTHIS THEN
      threadset[NTHIS].taskID := taskID;
      threadset[NTHIS].at := at;

      IF g # -1 THEN
         threadset[NTHIS].g := g + s;
      ELSE
         threadset[NTHIS].g := -1;
      END;

      threadset[NTHIS].a := a;
      threadset[NTHIS].s := s;

      INC(NTHIS)
   ELSE THREADSETOVR := TRUE;
      HLog.Str("Thread set overflow!"); HLog.Ln();
   END
END CreateThreadInstance; 
*)

PROCEDURE (8) CreateThreadInstance2();
BEGIN
   IF NTHIS < MAXTHIS THEN
      threadset[NTHIS].taskID := taskID;
      threadset[NTHIS].at := at;

      IF g # -1 THEN
         threadset[NTHIS].g := g + s;
      ELSE
         threadset[NTHIS].g := -1;
      END;

      threadset[NTHIS].a := a;
      threadset[NTHIS].s := s;

      INC(NTHIS)
   ELSE THREADSETOVR := TRUE;
   END
END CreateThreadInstance2; 

PROCEDURE (8) SystemInterpreter;
BEGIN
   WHILE pc # -1 DO
      opcode := program[pc].opcode;

      IF opcode = DISPATCH THEN
         IF taskset[program[pc].arg1].state = COMPLETE THEN (* optimization *)
            INC(pc)
         ELSE (* task has not yet completed *) 
            taskID := program[pc].arg1; at := pc + 1; g := program[pc].arg2; a := program[pc].arg3; s := referencestate;
            CreateThreadInstance2();
            pc := -1;
           END 
       ELSIF opcode = SCHEDULE THEN
        taskID := program[pc].arg1; deadline := program[pc].arg2; s := logicaltime;
        CreateTaskInstance2();
         INC(pc)
      ELSIF opcode = FUTURE THEN
         g := program[pc].arg1; a := program[pc].arg2; s := logicaltime;
         CreateTriggerBinding2();
         INC(pc)
       ELSIF opcode = RET THEN
         pc := -1
    ELSIF opcode = CALL THEN
         drivercode := drivers[program[pc].arg1];

         drivercode();

         INC(pc)
       ELSIF opcode = IDLE THEN
            taskID := -1; at := -1; g := program[pc].arg1; a := pc + 1; s := referencestate;
            CreateThreadInstance2();
         pc := -1;
      ELSIF opcode = FORK THEN
            taskID := -1; at := -1; g := 0; a := program[pc].arg1; s := logicaltime;
            CreateThreadInstance2();
         INC(pc)
      ELSIF opcode = JUMP THEN
         pc := program[pc].arg1
      ELSIF opcode = IFT THEN
         conditioncode := conditions[program[pc].arg1];
      
         (* branch := conditioncode(); not possible via optimization returns are on R0!*)
         branch := FALSE;
         IF branch THEN
            pc := program[pc].arg2
         ELSE
            pc := program[pc].arg3
         END
      ELSIF opcode = NOP THEN
         INC(pc)
     ELSE UNDEFINSTR := TRUE;

         pc := -1
      END
   END
END SystemInterpreter;

PROCEDURE (8)EventLoop;
BEGIN
   yieldtotasks := FALSE;

   WHILE ~yieldtotasks DO
      index := 0;
      hit := -1;

      WHILE index < NTHIS DO
         IF (threadset[index].taskID # -1) & (taskset[threadset[index].taskID].state = COMPLETE) THEN
            (* completed task in thread instance *)
            hit := index;
            index := NTHIS
         ELSE
            INC(index)
         END
      END;

      IF hit # -1 THEN (* there is a completed task in the thread set *)
         pc := threadset[hit].at;
         referencestate := threadset[hit].s;

         index := hit + 1;

         WHILE index < NTHIS DO (* remove enabled thread instance *)
            threadset[index - 1].taskID := threadset[index].taskID;
            threadset[index - 1].at := threadset[index].at;
            threadset[index - 1].g := threadset[index].g;
            threadset[index - 1].a := threadset[index].a;
            threadset[index - 1].s := threadset[index].s;

            INC(index)
         END;

         DEC(NTHIS);

         (* profiler *)
         SYSTEM.GET(OSCR, delta1);
         SystemInterpreter();
         SYSTEM.GET(OSCR, delta2);
            IF delta2 > delta1 THEN delta2 := delta2 - delta1; 
                IF (delays < 5000) THEN
                    scode := scode + delta2;
                    INC(delays);
                END;
            END;
      ELSE (* no completed task in the thread set *)
         index := 0;

         WHILE index < NTRGB DO
            IF triggerqueue[index].g = logicaltime THEN
               (* interrupt trigger enabled *)
               hit := index;
               index := NTRGB
            ELSE
               INC(index)
            END
         END;

         IF hit # -1 THEN (* there is an enabled trigger binding *)
            pc := triggerqueue[hit].a;
            referencestate := logicaltime;
            
            (* tracer: *)
            IF (pc = 0) & ( traceidx > 0) & (traceidx < MAXTRACE)   THEN  (* end of period *)
                traces[traceidx].time:= -2;
                INC(traceidx); traceidx := MAXTRACE;
            ELSIF (traceidx < MAXTRACE) THEN 
                traces[traceidx].time:= -1;
                traces[traceidx].from:= logicaltime;
                traces[traceidx].to:= pc;
                INC(traceidx);
            END;

            index := hit + 1;

            WHILE index < NTRGB DO (* remove enabled trigger binding *)
               triggerqueue[index - 1].g := triggerqueue[index].g;
               triggerqueue[index - 1].a := triggerqueue[index].a;

               INC(index)
            END;

            DEC(NTRGB);
            (* profile ecode *)
            SYSTEM.GET(OSCR, delta1);
            SystemInterpreter();
            SYSTEM.GET(OSCR, delta2);
            IF delta2 > delta1 THEN delta2 := delta2 - delta1; 
                IF (delaye < 5000) THEN
                    ecode := ecode + delta2;
                    INC(delaye);
                END;
            END;
            
        ELSE (* no completed tasks and no enabled trigger bindings *)
            index := 0;

            WHILE index < NTHIS DO
               IF threadset[index].g # -1 THEN (* there is a timeout *)
                  (* IF threadset[index].s > logicaltime THEN
                      IF threadset[index].g <= logicaltime + TIMEOVERFLOW THEN
                        (* timeout has expired *)
                        hit := index;
                        index := NTHIS
                     ELSE  (* timeout has not yet expired *)
                        INC(index)
                     END
                  ELS*)
                  IF threadset[index].g <= logicaltime THEN
                     (* timeout has expired *)
                     hit := index;
                     index := NTHIS
                  ELSE (* timeout has not yet expired *)
                     INC(index)
                  END
               ELSE (* no timeout set *)
                  INC(index)
               END
            END;

            IF hit # -1 THEN (* there is an enabled thread instance *)
               pc := threadset[hit].a;
               referencestate := threadset[hit].s;

               index := hit + 1;

               WHILE index < NTHIS DO (* remove enabled thread instance *)
                  threadset[index - 1].taskID := threadset[index].taskID;
                  threadset[index - 1].at := threadset[index].at;
                  threadset[index - 1].g := threadset[index].g;
                  threadset[index - 1].a := threadset[index].a;
                  threadset[index - 1].s := threadset[index].s;

                  INC(index)
               END;

               DEC(NTHIS);
               
               (* profiler *)
               SYSTEM.GET(OSCR, delta1);
               SystemInterpreter();
               SYSTEM.GET(OSCR, delta2);
                IF delta2 > delta1 THEN delta2 := delta2 - delta1; 
                    IF (delays < 5000) THEN
                        scode := scode + delta2;
                        INC(delays);
                    END;
                END;
            ELSE (* no completed tasks, no enabled trigger bindings, no enabled thread instances *)
               yieldtotasks := TRUE
            END;
         END;
      END;
   END;
END EventLoop;

PROCEDURE (8) TaskScheduler;
BEGIN
   IF NTHIS > 0 THEN (* schedule thread set *)
      hit := threadset[0].taskID;

   ELSIF NRT > 0 THEN  (* schedule task set, earliest deadline first only if there are non completed tasks*)
      index := 0;
      hit := -1;
      SYSTEM.GET(OSCR, delta1);
      WHILE index < NTKIS DO
         IF ((taskset[index].state # COMPLETE) & ((hit = -1) OR (taskset[index].deadline < taskset[hit].deadline))) THEN
               hit := index
        END;
         INC(index)
      END;
        SYSTEM.GET(OSCR, delta2);
        IF delta2 > delta1 THEN delta2 := delta2 - delta1; 
            IF (delayedf < 5000) THEN
            edf := edf + delta2;
            INC(delayedf);
            END;
        END;
    
   END
END TaskScheduler;

PROCEDURE (8) CopyContextFrom;
BEGIN
    IF runningtask = -1 THEN
        spsr := main.spsr;
        sp := main.sp;
        spsrfiq := main.spsrfiq;
        fpfiq := main.fpfiq;
        spfiq := main.spfiq
    ELSE 
        spsr := taskset[runningtask].context.spsr;
        sp := taskset[runningtask].context.sp;
        spsrfiq := taskset[runningtask].context.spsrfiq;
        fpfiq := taskset[runningtask].context.fpfiq;
        spfiq := taskset[runningtask].context.spfiq
    END;
END CopyContextFrom;

PROCEDURE (8) CopyContextTo;
BEGIN
    IF runningtask = -1 THEN
        main.spsr := spsr;
        main.sp := sp;
        main.spsrfiq := spsrfiq;
        main.fpfiq := fpfiq;
        main.spfiq := spfiq
    ELSE 
        taskset[runningtask].context.spsr := spsr;
        taskset[runningtask].context.sp := sp;
        taskset[runningtask].context.spsrfiq := spsrfiq;
        taskset[runningtask].context.fpfiq := fpfiq;
        taskset[runningtask].context.spfiq := spfiq
    END;
END CopyContextTo;

PROCEDURE InitTask(task: INTEGER); (* from IFQ Mode *)
BEGIN
    taskset[task].context.sp := SYSTEM.ADR(taskset[task]) + SIZE(taskset[task]);
    taskset[task].context.spfiq := SYSTEM.ADR(taskset[task]) + SIZE(taskset[task]) - 1024*4; 
END InitTask;

(* System machine is entered when a task is completed *)
PROCEDURE (8) SystemMachineCompletion();
BEGIN
    SYSTEM.LDPSR(21H, HKernel.SVCMode+ HKernel.FIQDisabled + HKernel.IRQDisabled);
    SYSTEM.PUT(GPSR, LED0);
    taskset[runningtask].deadline := -1;
    taskset[runningtask].state := COMPLETE;
    DEC(NRT);
    
      (* SMACHINE  *)
        SYSTEM.PUT(GPCR, LED0);
        SYSTEM.GET(OSCR, icfp);
        EventLoop();
        TaskScheduler();
        (* profiler *)
        SYSTEM.GET(OSCR, delta);
        IF delta > icfp THEN delta := delta - icfp; 
            IF (delay < 5000) THEN
                IF delta < min THEN min := delta END;
                mean := mean + delta;
                IF (delay > 1000) & (peak < delta) THEN peak := delta END;
                INC(delay);
            END;
        END;
        SYSTEM.PUT(GPSR, LED0);
        
        IF hit = -1 THEN  utilization := utilization + TIMESTEP END;
        IF (traceidx < MAXTRACE) THEN 
            traces[traceidx].time:= logicaltime;
            traces[traceidx].from:= runningtask;
            traces[traceidx].to:= hit;
            traces[traceidx].c:= TRUE;
            INC(traceidx);
        END;
        (* IF hit = 0 THEN SYSTEM.PUT(GPSR, LED2)
        ELSIF hit = 1 THEN SYSTEM.PUT(GPSR, LED3)
        END;
        IF runningtask = 0 THEN SYSTEM.PUT(GPCR, LED2)
        ELSIF runningtask = 1 THEN SYSTEM.PUT(GPCR, LED3)
        END; *)
        runningtask := hit;

    IF (runningtask #-1) & (taskset[runningtask].state = RELEASED) THEN
        taskset[runningtask].state := DISPATCHED;
        (* init task status *)
        (* INLINE OF: InitTask(runningtask); *)
                s := SIZE(taskset[runningtask]);
                a := SYSTEM.ADR(taskset[runningtask]) + s;
                taskset[runningtask].context.sp := a;
                a := a - 1024*4;
                taskset[runningtask].context.spfiq := a ; 
        SYSTEM.LDPSR(21H, HKernel.FIQMode+ HKernel.FIQDisabled + HKernel.IRQDisabled); 
        (* leave FIQ mode silently running the task code*)
        SYSTEM.SP := taskset[runningtask].context.spfiq;
        SYSTEM.LDPSR(21H, HKernel.SVCMode+ HKernel.FIQDisabled + HKernel.IRQDisabled); 
        SYSTEM.SP := taskset[runningtask].context.sp;
        localproc := taskset[runningtask].taskcode;
        SYSTEM.PUT(GPCR, LED0);
        SYSTEM.LDPSR(21H, HKernel.SVCMode); 
        localproc(); (* call coroutine *)
        SystemMachineCompletion(); 
        WHILE TRUE DO END;
    ELSE
        CopyContextFrom();
    END;
    IF runningtask = -1 THEN utilization := utilization + TIMESTEP END;
    (*restore context of the process to be run *)
    (*restore the running task *)
    SYSTEM.LDPSR(69H, spsr); (* SPSR := old *)
    SYSTEM.SP := sp;
    SYSTEM.CODE(0E8BD5FFFH); (* LDM         !SP , {R0 R1 R2 R3 R4 R5 R6 R7 R8 R9 RA RB FP LR } *)
    SYSTEM.LDPSR(21H, HKernel.FIQMode+ HKernel.FIQDisabled + HKernel.IRQDisabled); (* change mode all interrupts *)
    SYSTEM.FP := fpfiq;
    SYSTEM.SP := spfiq;
    SYSTEM.LDPSR(69H, spsrfiq); (* SPSR := old *)
    SYSTEM.PUT(GPCR, LED0);
    SYSTEM.CODE(0E1A0D00CH); (* MOV         SP ,FP *)
    SYSTEM.CODE(0E8BD5000H); (* LDM         !SP , {FP LR } *)
    SYSTEM.CODE(0E25EF004H); (* SUB  S      PC ,LR ,4*) (* does the return from interrupt !!!! *)
END SystemMachineCompletion;

PROCEDURE [14H] SystemMachine; (* Timer interrupt will trigger  the System Machine*)
    (* we change stack often, so do not use any local var !!! *)
BEGIN (* interrupts are disabled *)
    SYSTEM.CODE(0E92D5000H); (* STM         !SP , {FP LR } *)
    SYSTEM.CODE(0E28DC000H); (* ADD         FP ,SP ,0 *)
    SYSTEM.PUT(GPSR, LED1);
    
    (* save contex of running task *)
    (* 
    SYSTEM.LDPSR(21H, HKernel.SVCMode+ HKernel.FIQDisabled + HKernel.IRQDisabled); (* change mode all interrupts *)
    SYSTEM.CODE(0E92D5FFFH); (* STM         !SP , {R0 R1 R2 R3 R4 R5 R6 R7 R8 R9 RA RB FP LR } *)
    SYSTEM.STPSR(4, spsr);  (*store SPSR *)
    sp := SYSTEM.SP;
    SYSTEM.LDPSR(21H, HKernel.FIQMode+ HKernel.FIQDisabled + HKernel.IRQDisabled); (* change mode all interrupts *)
    *)
    SYSTEM.STPSR(4, spsrfiq);  (*store SPSR *)
    spfiq :=SYSTEM.SP;
    fpfiq := SYSTEM.FP;
    
    (* CopyContextTo(); *)
    
    SYSTEM.GET(ICFP, icfp);
    IF SYSTEM.AND(icfp, OSTimerBit) # 0 THEN
        SYSTEM.GET(OSMR0, icfp); SYSTEM.PUT(OSMR0, icfp + OSDELAY);
        SYSTEM.PUT(OSSR, 1); (* Clear Match *)
        logicaltime := (logicaltime + TIMESTEP) (* MOD TIMEOVERFLOW *);
        
        (* SMACHINE *)
        SYSTEM.PUT(GPCR, LED1);
        SYSTEM.GET(OSCR, icfp);
        EventLoop();
        TaskScheduler();
        (* profiler *)
        SYSTEM.GET(OSCR, delta);
        IF delta > icfp THEN delta := delta - icfp; 
            IF (delay < 5000) THEN
                IF delta < min THEN min := delta END;
                mean := mean + delta;
                IF (delay > 1000) & (peak < delta) THEN peak := delta END;
                INC(delay);
            END;
        END;
        IF hit = -1 THEN  utilization := utilization + TIMESTEP END;
        
        IF hit # runningtask THEN (* store the context only if needed *)
            SYSTEM.LDPSR(21H, HKernel.SVCMode+ HKernel.FIQDisabled + HKernel.IRQDisabled); 
            SYSTEM.CODE(0E92D5FFFH); (* STM         !SP , {R0 R1 R2 R3 R4 R5 R6 R7 R8 R9 RA RB FP LR } *)
            SYSTEM.STPSR(4, spsr);  (*store SPSR *)
            sp := SYSTEM.SP;
            SYSTEM.LDPSR(21H, HKernel.FIQMode+ HKernel.FIQDisabled + HKernel.IRQDisabled); 
            CopyContextTo();
            
            (* profiler *)
            IF (runningtask #-1) THEN INC(preemption) END;
            (* trace *)
            IF (traceidx < MAXTRACE) THEN 
                traces[traceidx].time:= logicaltime;
                traces[traceidx].from:= runningtask;
                traces[traceidx].to:= hit;
                traces[traceidx].c:= FALSE;
                INC(traceidx);
            END;
            (* IF hit = 0 THEN SYSTEM.PUT(GPSR, LED2)
            ELSIF hit = 1 THEN SYSTEM.PUT(GPSR, LED3)
            END;
            IF runningtask = 0 THEN SYSTEM.PUT(GPCR, LED2)
            ELSIF runningtask = 1 THEN SYSTEM.PUT(GPCR, LED3)
            END; *)
        END;    
        
        SYSTEM.PUT(GPSR, LED1);
    
        (* Start a new released TASK whould never return !!! *)   
        IF (hit #-1) & (taskset[hit].state = RELEASED) THEN
            runningtask := hit;
            taskset[runningtask].state := DISPATCHED;
            (* init task status *)
            (* INLINE OF: InitTask(runningtask); *)
                s := SIZE(taskset[runningtask]);
                a := SYSTEM.ADR(taskset[runningtask]) + s;
                taskset[runningtask].context.sp := a;
                a := a - 1024*4;
                taskset[runningtask].context.spfiq := a ; 
            (* leave FIQ mode silently running the task code*)
            SYSTEM.SP := taskset[runningtask].context.spfiq;
            SYSTEM.LDPSR(21H, HKernel.SVCMode+ HKernel.FIQDisabled + HKernel.IRQDisabled); 
            SYSTEM.SP := taskset[runningtask].context.sp;
            localproc := taskset[runningtask].taskcode;
            SYSTEM.PUT(GPCR, LED1);
            SYSTEM.LDPSR(21H, HKernel.SVCMode); 
            localproc(); (* call coroutine *)
            SystemMachineCompletion(); 
            WHILE TRUE DO END;
        ELSIF hit # runningtask THEN
            runningtask := hit;
            CopyContextFrom();
            SYSTEM.LDPSR(21H, HKernel.SVCMode+ HKernel.FIQDisabled + HKernel.IRQDisabled); 
            SYSTEM.LDPSR(69H, spsr); (* SPSR := old *)
            SYSTEM.SP := sp;
            SYSTEM.CODE(0E8BD5FFFH); (* LDM         !SP , {R0 R1 R2 R3 R4 R5 R6 R7 R8 R9 RA RB FP LR } *)
            SYSTEM.LDPSR(21H, HKernel.FIQMode+ HKernel.FIQDisabled + HKernel.IRQDisabled); 
        END;    
    
    END; 
    SYSTEM.FP := fpfiq; (* should not use any registers !!! r0 - rb *)
    SYSTEM.SP := spfiq; (* should not use any registers !!! r0 - rb *)
    SYSTEM.LDPSR(69H, spsrfiq); (* SPSR := old *) 
    SYSTEM.PUT(GPCR, LED1);
    
    SYSTEM.CODE(0E1A0D00CH); (* MOV         SP ,FP *)
    SYSTEM.CODE(0E8BD5000H); (* LDM         !SP , {FP LR } *)
END SystemMachine;

PROCEDURE SystemCode*(op, a1, a2, a3: INTEGER);
BEGIN
    program[NINS].opcode := op;
    program[NINS].arg1 := a1;
    program[NINS].arg2 := a2;
    program[NINS].arg3 := a3;
    INC(NINS);
END SystemCode; 

PROCEDURE SetSystemCodeFxp*(adr, a3: INTEGER);
BEGIN
    program[adr].arg3 := a3;
END SetSystemCodeFxp;   

PROCEDURE SetSystemCodeJump*(adr, a1: INTEGER);
BEGIN
    program[adr].opcode := JUMP;
    program[adr].arg1 := a1;
    HLog.Int(a1);
END SetSystemCodeJump;  

PROCEDURE ShowPrg*;
    VAR i: INTEGER;
BEGIN i := 0;
    WHILE (i < NINS) DO
        HLog.Int(i); HLog.Ch(":"); 
        HLog.Int(program[i].opcode); HLog.Ch(" "); HLog.Int(program[i].arg1);
        HLog.Ch(","); HLog.Int(program[i].arg2);HLog.Ch(","); HLog.Int(program[i].arg3); 
        HLog.Ln;
        INC(i);
    END
END ShowPrg;

PROCEDURE Start*;
BEGIN
    DisableInterrupt();
    HelyOS.StopSync;
    
    HKernel.LED(HKernel.LED0, FALSE);
    logicaltime := -TIMESTEP;
    
    InitTaskSet();
    NTKIS := NTSK;
    
    NTRGB := 0; (* number of trigger bindings *)
    CreateTriggerBinding(TIMESTEP, 0, logicaltime); (* creates an enabled trigger binding to address 0 *)
    
    NTHIS := 0; (* number of thread instances *)
    
    runningtask := -1;
    utilization := 0;
    peak := 0;  min := MAX(INTEGER); delay := 0; mean := 0;
    scode := 0;  ecode := 0; edf := 0; delays := 0; delaye := 0; delayedf := 0;
    preemption := 0;
    traceidx:= 0;
    ResetErrors();
    HLog.Str("Scheduler set to system machine"); HLog.Ln;
    HKernel.InstallHandler(SystemMachine, HKernel.FIQ);  (* interrupt handler *)  
    ReRouteOSTimer(); (* reroute timmer interrupt to SystemMachine handler *)
    EnableInterrupt();
END Start;

PROCEDURE Stop*;
BEGIN
    HelyOS.StopSync;
    HKernel.LED(HKernel.LED0, TRUE);
    HKernel.TimerHandler := NIL; (* interrupt handler *)
    SYSTEM.PUT(ICLR, 0); (* reroute all external interrupts to IRQ *)
    HKernel.EnableIRQ(); (* restart old HelyOS timer *) 
    HLog.Str("SystemMachine peak(us): "); HLog.Real(FLT(peak)/3.6864);
     HLog.Str(" min(us): "); HLog.Real(FLT(min)/3.6864); HLog.Str(" mean(us): ");
     HLog.Real(FLT(mean)/(3.6864*FLT(delay))); 
    HLog.Str(" P: "); HLog.Real(FLT(preemption)/FLT(logicaltime));
     HLog.Ln;
    HLog.Str("scode(us): ");  IF delays > 0 THEN HLog.Real(FLT(scode)/(3.6864*FLT(delays))); ELSE HLog.Int(0) END;
     HLog.Str(" ecode(us): "); IF delaye > 0 THEN HLog.Real(FLT(ecode)/(3.6864*FLT(delaye))); ELSE HLog.Int(0) END;
     HLog.Str(" edf(us): "); IF delayedf > 0 THEN HLog.Real(FLT(edf)/(3.6864*FLT(delayedf))); ELSE HLog.Int(0) END;
     HLog.Ln;
    HLog.Str("SystemMachine time, util, 1-%: "); HLog.Int(logicaltime); HLog.Ch(","); HLog.Int(utilization);
    HLog.Ch(","); HLog.Real(1.0 - FLT(utilization)/FLT(logicaltime)); HLog.Ln;
     ShowErrors();
END Stop;


(***************************************************************************)
(******************  TESTING CODE                                      *************************)
(***************************************************************************)

PROCEDURE (8) Driver0();
    VAR i, j:INTEGER;
BEGIN
    j := 0;
    WHILE j<5 DO
        INC(j);
    END;
END Driver0;

PROCEDURE (8) Driver1();
    VAR i, j:INTEGER;
BEGIN
    j := 0;
    WHILE j<5 DO
        INC(j);
    END;
END Driver1;

PROCEDURE Led*();
    VAR i, j:INTEGER;
BEGIN
    j := 0;
    WHILE j<5 DO
        FOR i := 0 TO LEDON DO HKernel.LED(HKernel.LED2, FALSE); END;
        FOR i := 0 TO LEDON DO HKernel.LED(HKernel.LED2, TRUE); END;
        INC(j);
    END;HKernel.LED(HKernel.LED2, FALSE); 
END Led;

PROCEDURE Led2*();
    VAR i, j :INTEGER;
BEGIN
    j := 0;
    WHILE j < 2 DO
        FOR i := 0 TO LEDON DO HKernel.LED(HKernel.LED3, FALSE); END;
        FOR i := 0 TO LEDON DO HKernel.LED(HKernel.LED3, TRUE); END;
        INC(j);
    END;HKernel.LED(HKernel.LED3, FALSE); 
END Led2;

PROCEDURE Led3*();
    VAR i, j :INTEGER;
BEGIN
    j := 0;
    WHILE j < 1 DO
        FOR i := 0 TO LEDON DO HKernel.LED(HKernel.LED3, FALSE); END;
        FOR i := 0 TO LEDON DO HKernel.LED(HKernel.LED3, TRUE); END;
        INC(j);
    END;HKernel.LED(HKernel.LED3, FALSE); 
END Led3;

PROCEDURE LedOn*();
    VAR i, j:INTEGER;
BEGIN HKernel.LED(HKernel.LED2, TRUE);
    j := 0;
    WHILE j<5 DO
        FOR i := 0 TO LEDON DO HKernel.LED(HKernel.LED2, TRUE); END;
        FOR i := 0 TO LEDON DO HKernel.LED(HKernel.LED2, TRUE); END;
        INC(j);
    END;HKernel.LED(HKernel.LED2, FALSE); 
END LedOn;

PROCEDURE LedOn2*();
    VAR i, j :INTEGER;
BEGIN HKernel.LED(HKernel.LED3, TRUE);
    j := 0;
    WHILE j < 2 DO
        FOR i := 0 TO LEDON DO HKernel.LED(HKernel.LED3, TRUE); END;
        FOR i := 0 TO LEDON DO HKernel.LED(HKernel.LED3, TRUE); END;
        INC(j);
    END;HKernel.LED(HKernel.LED3, FALSE); 
END LedOn2;

PROCEDURE On*();
    VAR i, j:INTEGER;
BEGIN 
    j := 0;
    WHILE j<5 DO
        FOR i := 0 TO LEDON DO  END;
        FOR i := 0 TO LEDON DO  END;
        INC(j);
    END; 
END On;

PROCEDURE On2*();
    VAR i, j :INTEGER;
BEGIN 
    j := 0;
    WHILE j < 2 DO
        FOR i := 0 TO LEDON DO END;
        FOR i := 0 TO LEDON DO END;
        INC(j);
    END; 
END On2;

PROCEDURE Setup(T: INTEGER; SCODE: BOOLEAN); (* we assume no preemption *)
    VAR i, a: INTEGER;
BEGIN
    LEDON := LEDCONST DIV T;
    FOR i := 0 TO T-1 DO 
        IF ~ODD(i) THEN tasks[i] := Led ELSE tasks[i] := Led2 END;
    END;
    NTSK := T;
    
    (* Setup Ecode *)
    NINS := 0;
    FOR i :=0 TO T DIV 2 -1 DO
        SystemCode(SCHEDULE, i, 2000, -1);
    END;
    FOR i :=T DIV 2 TO T-1 DO
        SystemCode(SCHEDULE, i, 1000, -1);
    END;
    IF SCODE THEN
        SystemCode(FUTURE, 1000, T+2 + T, -1);
        FOR i :=T-1  TO 0 BY -1 DO
            SystemCode(DISPATCH, i, 1000, T+1 + T);
        END;
    ELSE SystemCode(FUTURE, 1000, T+2, -1);
    END;
    SystemCode(RET, -1, -1, -1); 
    
    FOR i :=T DIV 2 TO T-1 DO
        SystemCode(SCHEDULE, i, 1000, -1);
    END;
    SystemCode(FUTURE, 1000, 0, -1);
    IF SCODE THEN
        a := NINS;
        FOR i := T-1 TO 0 BY -1 DO
            SystemCode(DISPATCH, i, 1000, a + T );
        END;
    END;
    SystemCode(RET, -1, -1, -1); 
    
    HLog.Str("Instructions: "); HLog.Int(NINS); HLog.Ln;
END Setup;

PROCEDURE Setup6(WCET, T: INTEGER; SCODE: BOOLEAN); (* we assume no preemption *)
    VAR i, a0, a1, a2, adr: INTEGER;
BEGIN
    LEDON := WCET;
    a0 := T DIV 4; a1 := a0*2; a2 := a0*3;
    FOR i := 0 TO T-1 DO 
        IF i < a0 THEN tasks[i] := Led ELSIF i <a1 THEN tasks[i] := Led2 ELSE tasks[i] := Led3 END;
    END;
    NTSK := T;
    
    (* Setup Ecode *)
    (* TIME 0 *)
    NINS := 0;
    FOR i :=0 TO a0-1 DO
        SystemCode(SCHEDULE, i, 60, -1);
    END;
    FOR i :=a0 TO a1-1 DO
        SystemCode(SCHEDULE, i, 30, -1);
    END;
    FOR i :=a1 TO a2-1 DO
        SystemCode(SCHEDULE, i, 20, -1);
    END;
    FOR i :=a2 TO T-1 DO
        SystemCode(SCHEDULE, i, 10, -1);
    END;
    adr := NINS;
    IF SCODE THEN
        SystemCode(FUTURE, 10, adr + T+2, -1);
        FOR i :=T-1  TO 0 BY -1 DO
            SystemCode(DISPATCH, i, 10, adr+T+1);
        END;
    ELSE SystemCode(FUTURE, 10, adr+2, -1);
    END;
    SystemCode(RET, -1, -1, -1); 
    (* TIME 10 *)
    FOR i :=a2 TO T-1 DO
        SystemCode(SCHEDULE, i, 10, -1);
    END;
    adr := NINS;
    IF SCODE THEN
        SystemCode(FUTURE, 10, adr + T+2, -1);
        FOR i :=T-1  TO 0 BY -1 DO
            SystemCode(DISPATCH, i, 10, adr + T+1);
        END;
    ELSE SystemCode(FUTURE, 10, adr+2, -1);
    END;
    SystemCode(RET, -1, -1, -1); 
    (* TIME 20 *)
    FOR i :=a1 TO a2-1 DO
        SystemCode(SCHEDULE, i, 20, -1);
    END;
    FOR i :=a2 TO T-1 DO
        SystemCode(SCHEDULE, i, 10, -1);
    END;
    adr := NINS;
    IF SCODE THEN
        SystemCode(FUTURE, 10, adr + T+2, -1);
        FOR i :=T-1  TO 0 BY -1 DO
            SystemCode(DISPATCH, i, 10, adr+T+1);
        END;
    ELSE SystemCode(FUTURE, 10, adr+2, -1);
    END;
    SystemCode(RET, -1, -1, -1); 
    (* TIME 30 *)
    FOR i :=a0 TO a1-1 DO
        SystemCode(SCHEDULE, i, 30, -1);
    END;
    FOR i :=a2 TO T-1 DO
        SystemCode(SCHEDULE, i, 10, -1);
    END;
    adr := NINS;
    IF SCODE THEN
        SystemCode(FUTURE, 10, adr + T+2, -1);
        FOR i :=T-1  TO 0 BY -1 DO
            SystemCode(DISPATCH, i, 10, adr+T+1);
        END;
    ELSE SystemCode(FUTURE, 10, adr+2, -1);
    END;
    SystemCode(RET, -1, -1, -1); 
    (* TIME 40 *)
    FOR i :=a1 TO a2-1 DO
        SystemCode(SCHEDULE, i, 20, -1);
    END;
    FOR i :=a2 TO T-1 DO
        SystemCode(SCHEDULE, i, 10, -1);
    END;
    adr := NINS;
    IF SCODE THEN
        SystemCode(FUTURE, 10, adr + T+2, -1);
        FOR i :=T-1  TO 0 BY -1 DO
            SystemCode(DISPATCH, i, 10, adr+T+1);
        END;
    ELSE SystemCode(FUTURE, 10, adr+2, -1);
    END;
    SystemCode(RET, -1, -1, -1); 
    (* TIME 50 *)
    FOR i :=a2 TO T-1 DO
        SystemCode(SCHEDULE, i, 10, -1);
    END;
    adr := NINS;
    IF SCODE THEN
        SystemCode(FUTURE, 10, 0, -1);
        FOR i :=T-1  TO 0 BY -1 DO
            SystemCode(DISPATCH, i, 10, adr+T+1);
        END;
    ELSE SystemCode(FUTURE, 10, 0, -1);
    END;
    SystemCode(RET, -1, -1, -1); 
    
HLog.Str("Instructions: "); HLog.Int(NINS); HLog.Ln;
END Setup6;

PROCEDURE EcodeEDF2*;
BEGIN
    Setup(2, FALSE);
END EcodeEDF2;

PROCEDURE EcodeEDF10*;
BEGIN
    Setup(10, FALSE);
END EcodeEDF10;

PROCEDURE EcodeEDF50*;
BEGIN
    Setup(50, FALSE);
END EcodeEDF50;

PROCEDURE EcodeEDF100*;
BEGIN Setup(100, FALSE);
END EcodeEDF100;

PROCEDURE Ecode6EDF4*;
BEGIN
    Setup6(1562, 4, FALSE);
END Ecode6EDF4;

PROCEDURE Ecode6EDF10*;
BEGIN
    Setup6(625, 10, FALSE);
END Ecode6EDF10;

PROCEDURE Ecode6EDF50*;
BEGIN
    Setup6(125, 50, FALSE);
END Ecode6EDF50;

PROCEDURE Ecode6EDF100*;
BEGIN Setup6(62, 100, FALSE);
END Ecode6EDF100;

PROCEDURE Ecode6T4*;
BEGIN
    Setup6(1562, 4, TRUE);
END Ecode6T4;

PROCEDURE Ecode6T10*;
BEGIN
    Setup6(625, 10, TRUE);
END Ecode6T10;

PROCEDURE Ecode6T50*;
BEGIN
    Setup6(125, 50, TRUE);
END Ecode6T50;

PROCEDURE Ecode6T100*;
BEGIN Setup6(62, 100, TRUE);
END Ecode6T100;

PROCEDURE Ecode6WEDF4*;
BEGIN
    Setup6(3000, 4, FALSE);
END Ecode6WEDF4;

PROCEDURE Ecode6WEDF10*;
BEGIN
    Setup6(1200, 10, FALSE);
END Ecode6WEDF10;

PROCEDURE Ecode6WEDF50*;
BEGIN
    Setup6(240, 50, FALSE);
END Ecode6WEDF50;

PROCEDURE Ecode6WEDF100*;
BEGIN Setup6(80, 100, FALSE);
END Ecode6WEDF100;

PROCEDURE Ecode6WT4*;
BEGIN
    Setup6(3000, 4, TRUE);
END Ecode6WT4;

PROCEDURE Ecode6WT10*;
BEGIN
    Setup6(1200, 10, TRUE);
END Ecode6WT10;

PROCEDURE Ecode6WT50*;
BEGIN
    Setup6(240, 50, TRUE);
END Ecode6WT50;

PROCEDURE Ecode6WT100*;
BEGIN Setup6(80, 100, TRUE);
END Ecode6WT100;

PROCEDURE Ecode6wT4*;
BEGIN
    Setup6(1562, 4, TRUE);
END Ecode6wT4;

PROCEDURE Ecode6wT10*;
BEGIN
    Setup6(625, 10, TRUE);
END Ecode6wT10;

PROCEDURE Ecode6wT50*;
BEGIN
    Setup6(80, 50, TRUE);
END Ecode6wT50;

PROCEDURE Ecode6wT100*;
BEGIN Setup6(35, 100, TRUE);
END Ecode6wT100;

PROCEDURE EcodeT2*;
BEGIN
    Setup(2, TRUE);
END EcodeT2;

PROCEDURE EcodeT10*;
BEGIN
    Setup(10, TRUE);
END EcodeT10;

PROCEDURE EcodeT50*;
BEGIN
    Setup(50, TRUE);
END EcodeT50;

PROCEDURE EcodeT100*;
BEGIN
    Setup(100, TRUE);
END EcodeT100;

PROCEDURE EcodeT100WP*;
BEGIN
    Setup(100, TRUE);
    LEDON := 3000;
END EcodeT100WP;


END FIQMachine4.

FIQMachine4.EcodeEDF2 peak(us):   7.3250E+00 min(us):   1.8991E+00 mean(us):   2.2025E+00 NO P!  4.4876E-01 7
FIQMachine4.EcodeEDF10 peak(us):   1.8446E+01 min(us):   4.0695E+00 mean(us):  4.6816E+00 NO P! 5.2478E-01 19
FIQMachine4.EcodeEDF50 peak(us):   1.1664E+02 min(us):   1.4377E+01 mean(us):   1.7555E+01 NO P! 5.7177E-01 79
FIQMachine4.EcodeEDF100 2.6937E+02 min(us):   2.7398E+01 mean(us):   3.4218E+01 NO P! 5.9825E-01 154

FIQMachine4.EcodeT2 peak(us):  8.9518E+00 min(us):   2.1701E+00 mean(us):   2.1970E+00 NO P! 4.4122E-01 10
FIQMachine4.EcodeT10 peak(us):   1.7090E+01 min(us):   2.1701E+00 mean(us):   3.3905E+00 NO P! 5.7338E-01 29
FIQMachine4.EcodeT50  peak(us):   8.3008E+01 min(us):   2.1701E+00 mean(us):   9.7482E+00 NO P! 5.7347E-01 129
FIQMachine4.EcodeT100 peak(us):   1.6656E+02 min(us):   2.1701E+00 mean(us):   1.7440E+01 NO P! 5.8386E-01 254

FIQMachine4.Ecode6EDF4 peak(us):   1.0851E+01 min(us):   2.4414E+00 mean(us):   3.2439E+00 pre: 134,0 3.4992E-01 24
FIQMachine4.Ecode6EDF10 peak(us):   1.9531E+01 min(us):   4.0690E+00 mean(us):   5.1763E+00 pre: 59,1 3.5105E-01 48
FIQMachine4.Ecode6EDF50 peak(us):  1.3808E+02 min(us): 1.4377E+01 mean(us):   1.9110E+01 pre: 191,4 4.0944E-01 168
FIQMachine4.Ecode6EDF100 peak(us): 2.8049E+02 min(us): 2.7398E+01 mean(us): 3.7423E+01 pre:172,20 5.880E-01 312 

(* WITH GENERATED SCODE *)
MachineTest.Ecode6EDF4 peak(us):   1.2750E+01 min(us):   2.1701E+00 mean(us):   3.4747E+00 pre: 154,0 3.5084E-01 51
MachineTest.Ecode6EDF10 peak(us):  1.6547E+01 min(us):   2.1701E+00 mean(us):   4.2442E+00 pre: 90,1 3.5088E-01 96
MachineTest.Ecode6EDF50 peak(us):   8.8162E+01 min(us):   1.8989E+00 mean(us):   7.6060E+00 pre: 87,5 3.6824E-01 326
MachineTest.Ecode6EDF100 peak(us): 1.6005E+02 min(us): 1.8989E+00 mean(us): 1.1293E+01 pre: 99,8 4.1754E-01 603   

FIQMachine4.Ecode6T4 peak(us):  1.1393E+01 min(us):   2.1701E+00 mean(us):   3.4201E+00 pre: 67,0 3.5118E-01 48
FIQMachine4.Ecode6T10 peak(us): 1.5734E+01 min(us):   2.1701E+00 mean(us):   4.6122E+00 pre: 95,0  3.5037E-01 108
FIQMachine4.Ecode6T50 peak(us):  8.1380E+01 min(us):   2.1701E+00 mean(us):   8.6236E+00 pre: 112,6 3.8451E-01 468
FIQMachine4.Ecode6T100 peak(us): 1.6683E+02 min(us):   2.1701E+00 mean(us):   1.0454E+01 pre: 78,16 4.0116E-01 912

FIQMachine4.Ecode6WEDF4 peak(us):   1.0851E+01 min(us):  2.4414E+00 mean(us): 3.3747E+00 pre: 256,1 8.1849E-01 24
FIQMachine4.Ecode6WEDF10 peak(us): 1.9531E+01 min(us): 4.0690E+00 mean(us): 5.5987E+00 pre: 236,2 9.1742E-01 48 
FIQMachine4.Ecode6WEDF50 peak(us):1.2505E+02 min(us):1.4377E+01 mean(us): 2.1200E+01 pre:292,10 9.4130E-01 168
FIQMachine4.Ecode6WEDF100 peak(us):2.7262E+02 min(us):2.7669E+01 mean(us):3.8420E+01 pre:528,47 7.814E-01 312

(* WITH GENERATED SCODE *)
MachineTest.Ecode6WEDF4 peak(us): 1.3021E+01 min(us):   2.1701E+00 mean(us):   3.5132E+00 pre: 356,1 8.1798E-01 51
MachineTest.Ecode6WEDF10 peak(us): 1.6819E+01 min(us):2.1701E+00 mean(us): 4.2927E+00 pre: 296,2 9.0041E-01 94   
MachineTest.Ecode6WEDF50 peak(us):  8.5449E+01 min(us):2.1701E+00 mean(us):9.2139E+00 pre:372,13 9.0063E-01 310
MachineTest.Ecode6WEDF100 peak(us):1.6819E+02 min(us):2.1701E+00 mean(us):1.3749E+01 pre:170,19 5.8496E-01 603 

FIQMachine4.Ecode6WT4 peak(us): 1.2207E+01 min(us):   2.1701E+00 mean(us):   3.5030E+00 pre: 320,0  8.6799E-01  48
FIQMachine4.Ecode6WT10 peak(us):.6276E+01 min(us):   2.1701E+00 mean(us):   4.1799E+00 pre:229,0  9.3217E-01 108  
FIQMachine4.Ecode6WT50 peak(us):8.6534E+01 min(us):  2.1701E+00 mean(us):   7.0617E+00 pre: 190,2 9.1784E-01 468  
FIQMachine4.Ecode6WT100 peak(us):1.6412E+02 min(us):2.7127E+00 mean(us):   9.7578E+00 pre: 142,4 5.6886E-01 912 

FIQMachine4.Ecode6wT4 peak(us): 1.2207E+01 min(us):   2.1701E+00 mean(us):   3.5030E+00 pre: 320,0  8.6799E-01  48
FIQMachine4.Ecode6wT10 peak(us):.6276E+01 min(us):   2.1701E+00 mean(us):   4.1799E+00 pre:229,0  9.3217E-01 108  
FIQMachine4.Ecode6wT50 peak(us):8.6534E+01 min(us):  2.1701E+00 mean(us):   7.0617E+00 pre: 190,2 9.1784E-01 468  
FIQMachine4.Ecode6wT100 peak(us):1.6412E+02 min(us):2.7127E+00 mean(us):   9.7578E+00 pre: 142,4 5.6886E-01 912 

Code Size for System Machine + IQ Handler 8832 bytes

*)

FIQMachine4.Ecode0
FIQMachine4.Ecode0np

FIQMachine4.TestEcode
FIQMachine4.TestRR

FIQMachine4.Led2

FIQMachine4.Start
FIQMachine4.Stop
